@@ -40,14 +40,6 @@
 #include "config.h"
 #include "eim.h"
 
-FCITX_EXPORT_API
-FcitxIMClass ime = {
-    FcitxChewingCreate,
-    FcitxChewingDestroy
-};
-FCITX_EXPORT_API
-int ABI_VERSION = FCITX_ABI_VERSION;
-
 CONFIG_DESC_DEFINE(GetFcitxChewingConfigDesc, "fcitx-chewing.desc")
 static int FcitxChewingGetRawCursorPos(char * str, int upos);
 static INPUT_RETURN_VALUE FcitxChewingGetCandWord(void* arg, FcitxCandidateWord* candWord);
@@ -55,16 +47,30 @@ static void FcitxChewingReloadConfig(void* arg);
 static boolean LoadChewingConfig(FcitxChewingConfig* fs);
 static void SaveChewingConfig(FcitxChewingConfig* fs);
 static void ConfigChewing(FcitxChewing* chewing);
+static void FcitxChewingOnClose(void* arg, FcitxIMCloseEventType event);
+static INPUT_RETURN_VALUE FcitxChewingKeyBlocker(void* arg, FcitxKeySym key, unsigned int state);
 
 typedef struct _ChewingCandWord {
     int index;
 } ChewingCandWord;
 
+FCITX_DEFINE_PLUGIN(fcitx_chewing, ime2, FcitxIMClass2) = {
+    FcitxChewingCreate,
+    FcitxChewingDestroy,
+    NULL,
+    NULL,
+    NULL,
+    NULL,
+    NULL,
+    NULL,
+};
+
+const FcitxHotkey FCITX_CHEWING_SHIFT_LEFT[2] = {{NULL, FcitxKey_Left, FcitxKeyState_Shift}, {NULL, FcitxKey_None, FcitxKeyState_None}};
+const FcitxHotkey FCITX_CHEWING_SHIFT_RIGHT[2] = {{NULL, FcitxKey_Right, FcitxKeyState_Shift}, {NULL, FcitxKey_None, FcitxKeyState_None}};
 const FcitxHotkey FCITX_CHEWING_UP[2] = {{NULL, FcitxKey_Up, FcitxKeyState_None}, {NULL, FcitxKey_None, FcitxKeyState_None}};
 const FcitxHotkey FCITX_CHEWING_DOWN[2] = {{NULL, FcitxKey_Down, FcitxKeyState_None}, {NULL, FcitxKey_None, FcitxKeyState_None}};
 const FcitxHotkey FCITX_CHEWING_PGUP[2] = {{NULL, FcitxKey_Page_Up, FcitxKeyState_None}, {NULL, FcitxKey_None, FcitxKeyState_None}};
 const FcitxHotkey FCITX_CHEWING_PGDN[2] = {{NULL, FcitxKey_Page_Down, FcitxKeyState_None}, {NULL, FcitxKey_None, FcitxKeyState_None}};
-int selKey[10] = {'1', '2', '3', '4', '5', '6', '7', '8', '9', '0'};
 
 const char *builtin_keymaps[] = {
     "KB_DEFAULT",
@@ -85,7 +91,6 @@ const char *builtin_keymaps[] = {
  * @param arg
  * @return successful or not
  **/
-__EXPORT_API
 void* FcitxChewingCreate(FcitxInstance* instance)
 {
     if (GetFcitxChewingConfigDesc() == NULL)
@@ -107,36 +112,37 @@ void* FcitxChewingCreate(FcitxInstance* instance)
     FcitxChewing* chewing = (FcitxChewing*) fcitx_utils_malloc0(sizeof(FcitxChewing));
     FcitxGlobalConfig* config = FcitxInstanceGetGlobalConfig(instance);
     FcitxInputState *input = FcitxInstanceGetInputState(instance);
-    FcitxCandidateWordSetChoose(FcitxInputStateGetCandidateList(input), DIGIT_STR_CHOOSE);
     
     bindtextdomain("fcitx-chewing", LOCALEDIR);
 
     chewing->context = chewing_new();
     ChewingContext * ctx = chewing->context;
     chewing->owner = instance;
-    chewing_set_ChiEngMode(ctx, CHINESE_MODE);
     chewing_set_maxChiSymbolLen(ctx, CHEWING_MAX_LEN);
     // chewing will crash without set page
     chewing_set_candPerPage(ctx, config->iMaxCandWord);
     FcitxCandidateWordSetPageSize(FcitxInputStateGetCandidateList(input), config->iMaxCandWord);
-    chewing_set_selKey(ctx, selKey, 10);
     LoadChewingConfig(&chewing->config);
     ConfigChewing(chewing);
 
-    FcitxInstanceRegisterIM(
+    FcitxIMIFace iface;
+    memset(&iface, 0, sizeof(FcitxIMIFace));
+    iface.Init = FcitxChewingInit;
+    iface.ResetIM = FcitxChewingReset;
+    iface.DoInput = FcitxChewingDoInput;
+    iface.GetCandWords = FcitxChewingGetCandWords;
+    iface.Save = NULL;
+    iface.ReloadConfig = FcitxChewingReloadConfig;
+    iface.OnClose = FcitxChewingOnClose;
+    iface.KeyBlocker = FcitxChewingKeyBlocker;
+
+    FcitxInstanceRegisterIMv2(
         instance,
         chewing,
         "chewing",
         _("Chewing"),
         "chewing",
-        FcitxChewingInit,
-        FcitxChewingReset,
-        FcitxChewingDoInput,
-        FcitxChewingGetCandWords,
-        NULL,
-        NULL,
-        FcitxChewingReloadConfig,
-        NULL,
+        iface,
         1,
         "zh_TW"
     );
@@ -151,27 +157,17 @@ void* FcitxChewingCreate(FcitxInstance* instance)
  * @param count count from XKeyEvent
  * @return INPUT_RETURN_VALUE
  **/
-__EXPORT_API
 INPUT_RETURN_VALUE FcitxChewingDoInput(void* arg, FcitxKeySym sym, unsigned int state)
 {
     FcitxChewing* chewing = (FcitxChewing*) arg;
-    FcitxInputState *input = FcitxInstanceGetInputState(chewing->owner);
     ChewingContext * ctx = chewing->context;
     int zuin_len;
-    FcitxCandidateWordList* candList = FcitxInputStateGetCandidateList(input);
 
-    if (FcitxCandidateWordGetListSize(candList) > 0) {
-        if (FcitxHotkeyIsHotKeyDigit(sym, state) || FcitxHotkeyIsHotKey(sym, state, FCITX_RIGHT) || FcitxHotkeyIsHotKey(sym, state, FCITX_LEFT))
-            return IRV_TO_PROCESS;
-        if (FcitxHotkeyIsHotKey(sym, state, FCITX_SPACE)) {
-            if (FcitxCandidateWordGoNextPage(candList))
-                return IRV_DISPLAY_MESSAGE;
-            else
-                return IRV_DO_NOTHING;
-        }
-    }
-
-    if (FcitxHotkeyIsHotKeySimple(sym, state)) {
+    if (FcitxHotkeyIsHotKey(sym, state, FCITX_SPACE)) {
+        chewing_handle_Space(ctx);
+    } else if (FcitxHotkeyIsHotKey(sym, state, FCITX_TAB)) {
+        chewing_handle_Tab(ctx);
+    } else if (FcitxHotkeyIsHotKeySimple(sym, state)) {
         int scan_code = (int) sym & 0xff;
         chewing_handle_Default(ctx, scan_code);
     } else if (FcitxHotkeyIsHotKey(sym, state, FCITX_BACKSPACE)) {
@@ -192,8 +188,6 @@ INPUT_RETURN_VALUE FcitxChewingDoInput(void* arg, FcitxKeySym sym, unsigned int
         chewing_handle_Del(ctx);
         if (chewing_buffer_Len(ctx) + zuin_len == 0)
             return IRV_CLEAN;
-    } else if (FcitxHotkeyIsHotKey(sym, state, FCITX_SPACE)) {
-        chewing_handle_Space(ctx);
     } else if (FcitxHotkeyIsHotKey(sym, state, FCITX_CHEWING_UP)) {
         chewing_handle_Up(ctx);
     } else if (FcitxHotkeyIsHotKey(sym, state, FCITX_CHEWING_DOWN)) {
@@ -206,6 +200,16 @@ INPUT_RETURN_VALUE FcitxChewingDoInput(void* arg, FcitxKeySym sym, unsigned int
         chewing_handle_Right(ctx);
     } else if (FcitxHotkeyIsHotKey(sym, state, FCITX_LEFT)) {
         chewing_handle_Left(ctx);
+    } else if (FcitxHotkeyIsHotKey(sym, state, FCITX_HOME)) {
+        chewing_handle_Home(ctx);
+    } else if (FcitxHotkeyIsHotKey(sym, state, FCITX_END)) {
+        chewing_handle_End(ctx);
+    } else if (FcitxHotkeyIsHotKey(sym, state, FCITX_SHIFT_SPACE)) {
+        chewing_handle_ShiftSpace(ctx);
+    } else if (FcitxHotkeyIsHotKey(sym, state, FCITX_CHEWING_SHIFT_LEFT)) {
+        chewing_handle_ShiftLeft(ctx);
+    } else if (FcitxHotkeyIsHotKey(sym, state, FCITX_CHEWING_SHIFT_RIGHT)) {
+        chewing_handle_ShiftRight(ctx);
     } else if (FcitxHotkeyIsHotKey(sym, state, FCITX_ENTER)) {
         chewing_handle_Enter(ctx);
     } else if (state == FcitxKeyState_Ctrl && FcitxHotkeyIsHotKeyDigit(sym, FcitxKeyState_None)) {
@@ -228,7 +232,6 @@ INPUT_RETURN_VALUE FcitxChewingDoInput(void* arg, FcitxKeySym sym, unsigned int
         return IRV_DISPLAY_CANDWORDS;
 }
 
-__EXPORT_API
 boolean FcitxChewingInit(void* arg)
 {
     FcitxChewing* chewing = (FcitxChewing*) arg;
@@ -238,11 +241,15 @@ boolean FcitxChewingInit(void* arg)
     return true;
 }
 
-__EXPORT_API
 void FcitxChewingReset(void* arg)
 {
     FcitxChewing* chewing = (FcitxChewing*) arg;
     chewing_Reset(chewing->context);
+
+    chewing_set_KBType( chewing->context, chewing_KBStr2Num(
+                (char *) builtin_keymaps[chewing->config.layout]));
+
+    chewing_set_ChiEngMode(chewing->context, CHINESE_MODE);
 #if 0
     FcitxUIStatus* puncStatus = FcitxUIGetStatusByName(chewing->owner, "punc");
     if (puncStatus) {
@@ -254,6 +261,36 @@ void FcitxChewingReset(void* arg)
 #endif
 }
 
+static boolean FcitxChewingPaging(void* arg, boolean prev)
+{
+    FcitxChewing* chewing = (FcitxChewing*) arg;
+    FcitxInputState *input = FcitxInstanceGetInputState(chewing->owner);
+    FcitxCandidateWordList* candList = FcitxInputStateGetCandidateList(input);
+
+    if (FcitxCandidateWordGetListSize(candList) > 0) {
+        return false;
+    }
+
+    if (prev) {
+        chewing_handle_Left(chewing->context);
+    } else {
+        chewing_handle_Right(chewing->context);
+    }
+
+    if (chewing_keystroke_CheckAbsorb(chewing->context)) {
+        return true;
+    }
+    return false;
+}
+
+static const char *builtin_selectkeys[] = {
+DIGIT_STR_CHOOSE,
+"asdfghjkl;",
+"asdfzxcv89",
+"asdfjkl789",
+"aoeuhtn789",
+"1234qweras",
+};
 
 /**
  * @brief function DoInput has done everything for us.
@@ -261,7 +298,6 @@ void FcitxChewingReset(void* arg)
  * @param searchMode
  * @return INPUT_RETURN_VALUE
  **/
-__EXPORT_API
 INPUT_RETURN_VALUE FcitxChewingGetCandWords(void* arg)
 {
     FcitxChewing* chewing = (FcitxChewing*) arg;
@@ -270,9 +306,18 @@ INPUT_RETURN_VALUE FcitxChewingGetCandWords(void* arg)
     FcitxMessages *clientPreedit = FcitxInputStateGetClientPreedit(input);
     ChewingContext * ctx = chewing->context;
     FcitxGlobalConfig* config = FcitxInstanceGetGlobalConfig(chewing->owner);
+    FcitxCandidateWordList* candList = FcitxInputStateGetCandidateList(input);
+
+    int selkey[10];
+    int i = 0;
+    for (i = 0; i < 10; i++) {
+        selkey[i] = builtin_selectkeys[chewing->config.selkey][i];
+    }
     
+    chewing_set_selKey(ctx, selkey, 10);
     chewing_set_candPerPage(ctx, config->iMaxCandWord);
-    FcitxCandidateWordSetPageSize(FcitxInputStateGetCandidateList(input), config->iMaxCandWord);
+    FcitxCandidateWordSetPageSize(candList, config->iMaxCandWord);
+    FcitxCandidateWordSetChoose(candList, builtin_selectkeys[chewing->config.selkey]);
 
     //clean up window asap
     FcitxInstanceCleanInputWindow(chewing->owner);
@@ -299,30 +344,52 @@ INPUT_RETURN_VALUE FcitxChewingGetCandWords(void* arg)
             cw.strExtra = NULL;
             cw.strWord = strdup(str);
             cw.wordType = MSG_OTHER;
-            FcitxCandidateWordAppend(FcitxInputStateGetCandidateList(input), &cw);
+            FcitxCandidateWordAppend(candList, &cw);
             chewing_free(str);
             index ++;
         }
+
+        if (FcitxCandidateWordGetListSize(candList) > 0) {
+            FcitxCandidateWordSetOverridePaging(
+                candList,
+                chewing_cand_CurrentPage(ctx) > 0,
+                chewing_cand_CurrentPage(ctx) + 1 < chewing_cand_TotalPage(ctx),
+                FcitxChewingPaging,
+                chewing,
+                NULL);
+        }
     }
-    
-    /* there is nothing */
-    if (strlen(zuin_str) == 0 && strlen(buf_str) == 0 && index == 0)
-        return IRV_DISPLAY_CANDWORDS;
 
-    // setup cursor
-    FcitxInputStateSetShowCursor(input, true);
-    int cur = chewing_cursor_Current(ctx);
-    FcitxLog(DEBUG, "cur: %d", cur);
-    int rcur = FcitxChewingGetRawCursorPos(buf_str, cur);
-    FcitxInputStateSetCursorPos(input, rcur);
-    FcitxInputStateSetClientCursorPos(input, rcur);
+    do {
+        /* there is nothing */
+        if (strlen(zuin_str) == 0 && strlen(buf_str) == 0 && index == 0)
+            break;
 
-    // insert zuin in the middle
-    char * half1 = strndup(buf_str, rcur);
-    char * half2 = strdup(buf_str + rcur);
-    FcitxMessagesAddMessageAtLast(msgPreedit, MSG_INPUT, "%s%s%s", half1, zuin_str, half2);
-    FcitxMessagesAddMessageAtLast(clientPreedit, MSG_INPUT, "%s%s%s", half1, zuin_str, half2);
-    chewing_free(buf_str); chewing_free(zuin_str); free(half1); free(half2);
+        // setup cursor
+        FcitxInputStateSetShowCursor(input, true);
+        int cur = chewing_cursor_Current(ctx);
+        FcitxLog(DEBUG, "cur: %d", cur);
+        int rcur = FcitxChewingGetRawCursorPos(buf_str, cur);
+        FcitxInputStateSetCursorPos(input, rcur);
+        FcitxInputStateSetClientCursorPos(input, rcur);
+
+        // insert zuin in the middle
+        char * half1 = strndup(buf_str, rcur);
+        char * half2 = strdup(buf_str + rcur);
+
+        FcitxMessagesAddMessageAtLast(msgPreedit, MSG_INPUT, "%s", half1);
+        FcitxMessagesAddMessageAtLast(msgPreedit, MSG_CODE, "%s", zuin_str);
+        FcitxMessagesAddMessageAtLast(msgPreedit, MSG_INPUT, "%s", half2);
+
+        FcitxMessagesAddMessageAtLast(clientPreedit, MSG_OTHER, "%s", half1);
+        FcitxMessagesAddMessageAtLast(clientPreedit, MSG_HIGHLIGHT | MSG_DONOT_COMMIT_WHEN_UNFOCUS, "%s", zuin_str);
+        FcitxMessagesAddMessageAtLast(clientPreedit, MSG_OTHER, "%s", half2);
+        free(half1);
+        free(half2);
+    } while(0);
+
+    chewing_free(buf_str);
+    chewing_free(zuin_str);
 
     return IRV_DISPLAY_CANDWORDS;
 }
@@ -351,7 +418,7 @@ INPUT_RETURN_VALUE FcitxChewingGetCandWord(void* arg, FcitxCandidateWord* candWo
         }
         lastPage = chewing_cand_CurrentPage(chewing->context);
     }
-    chewing_handle_Default( chewing->context, selKey[off] );
+    chewing_handle_Default( chewing->context, builtin_selectkeys[chewing->config.selkey][off] );
     
     if (chewing_keystroke_CheckAbsorb(chewing->context)) {
         return IRV_DISPLAY_CANDWORDS;
@@ -386,7 +453,6 @@ static int FcitxChewingGetRawCursorPos(char * str, int upos)
  *
  * @return int
  **/
-__EXPORT_API
 void FcitxChewingDestroy(void* arg)
 {
     FcitxChewing* chewing = (FcitxChewing*) arg;
@@ -401,6 +467,54 @@ void FcitxChewingReloadConfig(void* arg) {
     ConfigChewing(chewing);
 }
 
+void FcitxChewingOnClose(void* arg, FcitxIMCloseEventType event)
+{
+    FcitxChewing* chewing = (FcitxChewing*) arg;
+    ChewingContext* ctx = chewing->context;
+    if (event == CET_LostFocus || event == CET_ChangeByInactivate) {
+        chewing_handle_Enter(ctx);
+        if (event == CET_ChangeByInactivate) {
+            if (chewing_commit_Check(ctx)) {
+                char* str = chewing_commit_String(ctx);
+                FcitxInputContext* ic = FcitxInstanceGetCurrentIC(chewing->owner);
+                FcitxInstanceCommitString(chewing->owner, ic, str);
+                chewing_free(str);
+            } else {
+                char * buf_str = chewing_buffer_String(ctx);
+                do {
+                    /* there is nothing */
+                    if (strlen(buf_str) == 0)
+                        break;
+                    FcitxInputContext* ic = FcitxInstanceGetCurrentIC(chewing->owner);
+                    FcitxInstanceCommitString(chewing->owner, ic, buf_str);
+                } while(0);
+                chewing_free(buf_str);
+            }
+        }
+        else {
+            FcitxInstanceResetInput(chewing->owner);
+        }
+    }
+
+}
+
+INPUT_RETURN_VALUE FcitxChewingKeyBlocker(void* arg, FcitxKeySym key, unsigned int state)
+{
+    FcitxChewing* chewing = (FcitxChewing*) arg;
+    FcitxInputState *input = FcitxInstanceGetInputState(chewing->owner);
+    FcitxCandidateWordList* candList = FcitxInputStateGetCandidateList(input);
+    if (FcitxCandidateWordGetListSize(candList) > 0
+        && (FcitxHotkeyIsHotKeySimple(key, state)
+        || FcitxHotkeyIsHotkeyCursorMove(key, state)
+        || FcitxHotkeyIsHotKey(key, state, FCITX_SHIFT_SPACE)
+        || FcitxHotkeyIsHotKey(key, state, FCITX_TAB)
+        || FcitxHotkeyIsHotKey(key, state, FCITX_SHIFT_ENTER)
+        ))
+        return IRV_DO_NOTHING;
+    else
+        return IRV_TO_PROCESS;
+}
+
 boolean LoadChewingConfig(FcitxChewingConfig* fs)
 {
     FcitxConfigFileDesc *configDesc = GetFcitxChewingConfigDesc();
@@ -435,10 +549,9 @@ void SaveChewingConfig(FcitxChewingConfig* fc)
 void ConfigChewing(FcitxChewing* chewing)
 {
     ChewingContext* ctx = chewing->context;
-    chewing_set_KBType( ctx, chewing_KBStr2Num( 
-                (char *) builtin_keymaps[chewing->config.layout]));
     chewing_set_addPhraseDirection( ctx, chewing->config.bAddPhraseForward ? 0 : 1 );
     chewing_set_phraseChoiceRearward( ctx, chewing->config.bChoiceBackward ? 1 : 0 );
     chewing_set_autoShiftCur( ctx, chewing->config.bAutoShiftCursor ? 1 : 0 );
     chewing_set_spaceAsSelection( ctx, chewing->config.bSpaceAsSelection ? 1 : 0 );
+    chewing_set_escCleanAllBuf( ctx, 1 );
 }
